.doc-labels {
    display: none;
}

.doc-class>.doc-heading {
    margin-top: 0.5em;
    border-top: 3px #33333333 solid;
    background-color: var(--md-code-bg-color);
    /* padding-top: 0.5em; */
}

.doc-class {
    margin-bottom: 1em;
    border-bottom: 1px #33333333 solid;
}


.doc-function>.doc-heading {
    margin-top: 0.5em;
    border-left: 3px #33333333 solid;
    background-color: var(--md-code-bg-color);
    /* padding-top: 0.5em; */
}


.doc-attribute>h3,
.doc-function>h3 {
    font-size: 1.0rem;
}

.doc-contents>details.quote>summary {
    background-color: #FFFFFF;
    opacity: 0.6;
}

.doc-contents>details.quote {
    border-color: #33333333;
}

/* attributes, functions, ... */

.doc-attribute-tag {
    margin-left: 5px;
    font-size: 0.6em;
    font-style: italic;
    padding: 0.1em;
    color: green;
    background-color: rgb(220, 251, 220);
}

.doc-function-tag {
    margin-left: 5px;
    font-size: 0.6em;
    font-style: italic;
    padding: 0.1em;

    color: coral;
    background-color: rgb(251, 235, 220);
}

.doc-class-tag {
    margin-left: 5px;
    font-size: 0.6em;
    font-style: italic;
    padding: 0.1em;
    color: red;
    background-color: rgb(251, 220, 220);
}

/* bold name of attribute, function, ... */
.doc-attribute code:first-child {
    font-weight: bold;
}

.doc-function code:first-child {
    font-weight: bold;
}

.doc .md-typeset__table,
.doc .md-typeset__table table {
    display: table !important;
    width: inherit;
}

.md-typeset table:not([class]) th {
    padding-top: 0;
    padding-bottom: 0;
}